/*
	PSEmu Plugin

	Plugin Type: GPU
	Plugin Name: Null Driver
	Description: All drawing is not performed and also nothing is displayed

	(C)1997-1998 Vision Thing Software Group

	All rights reserved. Using this source or part of it without permission
	is highly prohibited.

    Changes, bug reports - plugin@psemu.com
*/

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <ddraw.h>
#define GPU_INTERNALS_DEF
#include "gpu_i.h"
#include "resource.h"
#include "CnfDlg.h"
#include "primitive drawing.h"
#include "PSEmu PlugIn Defs.h"
#include "AbtDlg.h"
#include "syslog.h"

// PPDK developer must change libraryName field and can change revision and build

const unsigned char version = PLUGIN_VERSION;	// do not touch - library for PSEmu 1.x

// it is up to developer but values must be in range 0-255

const unsigned char revision =	0;
const unsigned char build =		3;

// to obtain library name for your plugin, mail: plugin@psemu.com
// this must be unique, and only we can provide this
static char *libraryName =		"Zen OpenGL Driver";

HWND hwnd;
HDC hDC;
HGLRC hRC;
float scalarDispWidth,scalarDispHeight;
float scalarDrawWidth,scalarDrawHeight;


// driver dependant variables
GpuConfS gpuConfig;

static long initGPU = 0;
static long ScreenOpened = 0;

#define RED(x) (x & 0xff)
#define BLUE(x) ((x>>16) & 0xff)
#define GREEN(x) ((x>>8) & 0xff)

#define COLOR(x) (x & 0xffffff)

// macros for easy access to packet information
#define GPUCOMMAND(x) ((x>>24) & 0xff)

// memory image of the PSX vram 
unsigned char psxVub[1024*520*2];
signed char *psxVsb;
unsigned short *psxVuw;
signed short *psxVsw;
unsigned long *psxVul;
signed long *psxVsl;

long flip;

// internal GPU

static long GPUdataRet;
long GPUstatusRet;

static unsigned long gpuData[100];
static unsigned char gpuCommand = 0;
static long gpuDataC = 0;
static long gpuDataP = 0;

static long tWinX;
static long tWinY;
static long tWinW;
static long tWinH;

long drawingLines;

long dispWidths[4] = {256,320,512,640};
long dispPosX=0;
long dispPosY=0;
long dispVer = 0;
long dispHor = 0;
long dispHorNew;
long dispVerNew;
long dispColor = 0;
long dispColorNew;
long dispLace = 0;
long dispLaceNew;
long imageTransfer;
long drawLace;

short imTYc,imTXc,imTY,imTX;
long imSize;
short imageX0,imageX1;
short imageY0,imageY1;

unsigned short textBuf[512*512];
long newTextX0,newTextX1,newTextX2,newTextX3;
long newTextY0,newTextY1,newTextY2,newTextY3;

long gpuDataX;


char * CALLBACK PSEgetLibName(void)
{
	return libraryName;
}

unsigned long CALLBACK PSEgetLibType(void)
{
	return  PSE_LT_GPU;
}

unsigned long CALLBACK PSEgetLibVersion(void)
{
	return version<<16|revision<<8|build;
}




long CALLBACK GPUinit()
{
	HKEY key;
	DWORD type;
	DWORD temp;
	DWORD size;
	// initialise device here ... 
	// possibly configure the color format, etc.
	// reading registry settings
	RegCreateKey(HKEY_CURRENT_USER,"Software\\Vision Thing\\PSEmu Pro\\GPU\\ZenOpenGL",&key);
	  //Fullscreen
	size = 4;
	if(RegQueryValueEx(key,"FullScreen",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=TRUE;
		RegSetValueEx(key,"FullScreen",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bFullscreen=(BOOL)temp;
	  //Bilinear Filtering
	size = 4;
	if(RegQueryValueEx(key,"Bilinear",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=TRUE;
		RegSetValueEx(key,"Bilinear",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bBilinear=(BOOL)temp;
	  //Texture Half Resolution
	size = 4;
	if(RegQueryValueEx(key,"TextureHalf",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=TRUE;
		RegSetValueEx(key,"TextureHalf",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bTextureHalf=(BOOL)temp;
	  //Max Textures
	size = 4;
	if(RegQueryValueEx(key,"MaxTextures",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=32;
		RegSetValueEx(key,"MaxTextures",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.nMaxTextures=temp;
	  //Dump VRAM
	size = 4;
	if(RegQueryValueEx(key,"DumpVRAM",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=FALSE;
		RegSetValueEx(key,"DumpVRAM",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bDumpVRAM=temp;
	  //Wire Frame
	size = 4;
	if(RegQueryValueEx(key,"WireFrame",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=FALSE;
		RegSetValueEx(key,"WireFrame",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bWireFrame=temp;
	  //Antialias
	size = 4;
	if(RegQueryValueEx(key,"Antialias",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=FALSE;
		RegSetValueEx(key,"Antialias",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bAntialias=temp;
	  //Clear screen
	size = 4;
	if(RegQueryValueEx(key,"ClearScreen",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=FALSE;
		RegSetValueEx(key,"ClearScreen",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bClearScreen=temp;
	  //Check Palette
	size = 4;
	if(RegQueryValueEx(key,"CheckPalette",0,&type,(LPBYTE)&temp,&size)!=ERROR_SUCCESS){
		temp=FALSE;
		RegSetValueEx(key,"CheckPalette",0,REG_DWORD,(CONST BYTE *)&temp,4);
	}
	gpuConfig.bCheckPalette=temp;
	  //Close the key
	RegCloseKey(key);

	// mapping the VRAM
	psxVsb=(signed char *)psxVub;
	psxVsw=(signed short *)psxVub;
	psxVsl=(signed long *)psxVub;
	psxVuw=(unsigned short *)psxVub;
	psxVul=(unsigned long *)psxVub;

	GPUstatusRet = 0x74000000;
	return PSE_ERR_SUCCESS;
}


long CALLBACK GPUshutdown()
{
	if(initGPU==1)
	{
		initGPU=0;
	}
	return 0;
}

BOOL isFirst=TRUE;

void DoGLInit(){
	int i;
    PIXELFORMATDESCRIPTOR pfd={
        sizeof(PIXELFORMATDESCRIPTOR),
        1,
        PFD_DRAW_TO_WINDOW|
		  PFD_SUPPORT_OPENGL
		  |PFD_DOUBLEBUFFER
		,PFD_TYPE_RGBA,
        16,
        0,0,0,0,0,0,
        0,0,0,0,0,0,0,
        16,
        0,0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
	hDC=GetDC(hwnd);
    int pixelFormat = ChoosePixelFormat(hDC,&pfd);
    BOOL success = SetPixelFormat(hDC,pixelFormat,&pfd);
    hRC=wglCreateContext(hDC);
	if(hRC==NULL)
    {
        // mode set failed;
		ChangeDisplaySettings(NULL, CDS_FULLSCREEN);
		AfxMessageBox("Unable to create OpenGL Render Context");
		return;// -1;
    }
    wglMakeCurrent(hDC,hRC);

	glViewport(0,0,640,480);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glScalef(1.0f,-1.0f,1.0f);
	glOrtho(0.0,639.0,0.0,479.0,1.0,-1.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glMatrixMode(GL_TEXTURE);
	glLoadIdentity();
	glScalef(1.0f/255.0f,1.0f/255.0f,1.0f);

	glEnable(GL_CLIP_PLANE0);
	glEnable(GL_CLIP_PLANE1);
	glEnable(GL_BLEND);
	glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
	if(gpuConfig.bAntialias){
		glEnable(GL_LINE_SMOOTH);
		glEnable(GL_POLYGON_SMOOTH);
	}
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);

	glEnable(GL_TEXTURE_2D);
	for(i=0;i<gpuConfig.nMaxTextures;i++){
		texture[i].textAddrX=0;
		texture[i].textAddrX=0;
		texture[i].textTP=0;
		texture[i].clutP=nullclutP;
		texture[i].Update=FALSE;
		glGenTextures(1,&texture[i].id);
		glBindTexture(GL_TEXTURE_2D,texture[i].id);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);
		if(gpuConfig.bBilinear){
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		}else{
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		}
		glTexEnvi(GL_TEXTURE_ENV,GL_TEXTURE_ENV_MODE,GL_MODULATE);
		if(gpuConfig.bTextureHalf){
			glTexImage2D(GL_TEXTURE_2D,0,4,128,128,0,GL_RGBA,GL_UNSIGNED_BYTE,image);
		}else{
			glTexImage2D(GL_TEXTURE_2D,0,4,256,256,0,GL_RGBA,GL_UNSIGNED_BYTE,image);
		}
	}
	glGenTextures(1,&nullid);
	glBindTexture(GL_TEXTURE_2D,nullid);
	glTexImage2D(GL_TEXTURE_2D,0,4,0,0,0,GL_RGBA,GL_UNSIGNED_BYTE,0);

	glClearColor(0.0f,0.0f,0.0f,0.0f);
	glClear(GL_COLOR_BUFFER_BIT);

	if(gpuConfig.bWireFrame)
		glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);

	torgba[0]=0x00000000;
	for(i=1;i<65536;i++){
		torgba[i] =(i&0x001f)<<3;
		torgba[i]|=(i&0x03e0)<<6;
		torgba[i]|=(i&0x7c00)<<9;
		torgba[i]|=0xff000000;
	}

	syslog(0,(char*)glGetString(GL_VENDOR));
	syslog(0,(char*)glGetString(GL_RENDERER));
	syslog(0,(char*)glGetString(GL_VERSION));
	syslog(0,(char*)glGetString(GL_EXTENSIONS));
	isFirst=FALSE;
}

long CALLBACK GPUopen(HWND hwndGPU)
{
	// hwndGPU is a handle of window that PSEmu opens for emulation
    DEVMODE devmode;

	if(gpuConfig.bFullscreen){
		devmode.dmSize = sizeof(devmode);
	    devmode.dmBitsPerPel = 16;
	    devmode.dmPelsWidth = 640;
	    devmode.dmPelsHeight = 480;
	    devmode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;

	    if(ChangeDisplaySettings(&devmode, CDS_FULLSCREEN)!=DISP_CHANGE_SUCCESSFUL){
			// mode set failed;
			AfxMessageBox("Unable to change display resolution");
			return -1;
	    }
	}
	hwnd=hwndGPU;

    SetWindowPos(hwnd,HWND_TOP,
		0,0,640,480,0);

	ScreenOpened=1;
	return 0;
}

long CALLBACK GPUclose()
{
	// close the screen here
	// something like 'close()'
	if(ScreenOpened==1)
	{
		// add here screen closing routines
		wglDeleteContext(hRC);
		ReleaseDC(hwnd,hDC);
		if(gpuConfig.bFullscreen)ChangeDisplaySettings(NULL, CDS_FULLSCREEN);
		ScreenOpened=0;
		if(gpuConfig.bDumpVRAM==TRUE){
			FILE *out=fopen("vdump.raw","wb");
			fwrite(psxVub,2,1024*512,out);
			fclose(out);
		}
	}
	return 0;
}

BOOL db=FALSE;

void updateDisplay(void)
{
	// update Display on change position in VRAM or on VSync
	if(isFirst==TRUE){
		DoGLInit();
	}
	glFlush();
	SwapBuffers(hDC);
	if(gpuConfig.bClearScreen){
		glClearColor(0.33f,0.33f,0.33f,1.0f);
		glClear(GL_COLOR_BUFFER_BIT);
	}
	//syslog(0,"UPDATING DISPLAY");
}

void updateDisplayIfChanged(void)
{
	if ((dispVer == dispVerNew) && (dispHor == dispHorNew) && (dispColor == dispColorNew) && (dispLace == dispLaceNew))
		return; // do not perform display update.

	//glClearColor(0.0f,0.0f,0.0f,0.0f);
	//glClear(GL_COLOR_BUFFER_BIT);

	dispVer = dispVerNew;
	dispHor = dispHorNew;
	dispColor = dispColorNew;
	dispLace = dispLaceNew;

	updateDisplay();
	if(isFirst==FALSE&&dispHor>0&&dispVer>0){
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glScalef(1.0f,-1.0f,1.0f);
		glOrtho(0.0,(double)(dispHor-1),0.0,(double)(dispVer-1),1.0,-1.0);
	}

}

// update lace is called every VSync
void CALLBACK GPUupdateLace(void)
{
	__asm push edi
	if(dispLace)
	{
		drawingLines^=1;
		if(dispHor>0 && dispVer>0)
		{
			//syslog(0,"UPDATING LACE");
			updateDisplay();
		}
	}
	__asm pop edi
}


// process read request from GPU status register
unsigned long CALLBACK GPUreadStatus(void)
{
	// return the status of the GPU
	/*
	- gp1 - GPU Status
          Mask
        04000000 - 1:Idle 0:Busy
        10000000 - 1:Ready 0:Not ready , to receive commands.
        80000000 - GPU is drawing 1:Odd 0:Even lines, in interlaced mode.
	*/
	//just'0x74000000' for now ! (idle, ready, even lines)
	if (drawingLines==1)
		return GPUstatusRet;
	else
		return GPUstatusRet|0x80000000;
	//return 0x74000000;
}

short dispOffset;

// processes data send to GPU status register
// these are always single packet commands.
void CALLBACK GPUwriteStatus(unsigned long gdata)
{
	switch((gdata>>24)&0xff)
	{
	case 0x04:
		//syslog(77," GPU_S: TR_MODE: %x\n",gdata&0xffffff);
		if((gdata&0xffffff)==0) imageTransfer=0;
		if((gdata&0xffffff)==2) imageTransfer=3;
		return;
   
	case 0x05:
	    dispPosY = (gdata>>10)&0x3ff;
		dispPosX = gdata & 0x3ff;
		//syslog(77," GPU_S: DISPLAY SET: X=%d Y=%d\n",dispPosX,dispPosY);
		if (dispLace==0) updateDisplay();

		dispOffset=-(short)(dispVer-dispPosY);

		return;
	case 0x08:
		dispHorNew = dispWidths[gdata&0x3];
		if(gdata&0x40 && dispHorNew == 320) dispHorNew=384;
		if(gdata&0x40 && dispHorNew == 256) dispHorNew=352;
		if (gdata&0x04) dispVerNew = 480;
		else dispVerNew = 240;
		dispColorNew = (gdata>>4)&0x1; // if 1 - TrueColor
		dispLaceNew = (gdata>>5)&0x01; // if 1 - Interlace
		//syslog(77,"!GPU! DISPLAY SET: W=%d Wo=%d H=%d TRUE=%d LACE=%d\n",dispHorNew,dispWidths[gdata&0x3],dispVerNew,dispColorNew,dispLaceNew);
		if(dispVerNew==480 && dispLaceNew==1) drawLace = 1;
		else drawLace = 0;

		dispOffset=-(short)(dispVer-dispPosY);

		updateDisplayIfChanged();
		return;
	case 0x10: // ask about GPU version
		GPUdataRet=0x2; // set GPU version to 2
		return;

	default:
		syslog(77,"STATUS=%08x\n",gdata);
		return;
	}
	return;
}

unsigned long CALLBACK GPUreadData(void)
{
	if(imageTransfer==2)
	{
		// ****
		//imageTransfer = 0;
		
		// image transfer from VRAM
		if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
		{
			GPUdataRet=psxVul[imTY*512+imTX/2];
			//syslog(11,"RD: %08x",GPUdataRet);
		}
		imTX+=2;
		imTXc-=2;
		if(imTXc<=0)
		{
			imTX=imageX0;
			imTXc=imageX1;
			imTYc--;
			imTY++;
		}
		imSize--;          
		if(imSize <= 0)
		{
			GPUstatusRet&=0xf7ffffff;
			imageTransfer=0;
		}
	}
	return GPUdataRet;
}

// processes data send to GPU data register
// user should not take care about that
void CALLBACK GPUwriteData(unsigned long gdata)
{
	unsigned char command;
	GPUdataRet=gdata;
	//syslog(0,"WD: %08x",gdata);
	// here follows transfer to vram
	if((imageTransfer & 1) == 1)
	{
		// image transfer to VRAM
		if (imTYc>0 && imTXc>0)
		{
			if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
				psxVuw[(imTY<<10)+imTX]=(unsigned short)gdata;
			imTX++;
			imTXc--;
			if(imTXc<=0)
			{
				imTX=(imageX0);
				imTXc=(imageX1);
				imTYc--;
				imTY++;
			}
			if(imTYc <= 0){
				imageTransfer=0;
				return;
			}
     
			if ((imTY>=0) && (imTY<512) && (imTX>=0) && (imTX<1024))
				psxVuw[(imTY<<10)+imTX]=(unsigned short)(gdata>>16);

			imTX++;
			imTXc--;

			if(imTXc<=0)
			{
				imTX=(imageX0);
				imTXc=(imageX1);
				imTYc--;
				imTY++;
			}
     
			if(imTYc <= 0) 
				imageTransfer=0;
     
			return;
		}
		else
			imageTransfer=0;
	}
 
	if(gpuDataC == 0)
	{
		command = (unsigned char) (gdata>>24) & 0xff;
		if (primTableC[command])
		{
			gpuDataC = primTableC[command];
			gpuCommand = command;
			gpuData[0] = gdata;
			gpuDataP = 1;
		}
		else 
		{
			if (gdata)
			{
				syslog(77,"!GPU! ERROR: COMMAND? %02x, %06x\n",command,gdata);
			}
			return;      
		}
	}
	else
	{
		gpuData[gpuDataP] = gdata;
		gpuDataP++;
	}
	if(gpuDataP == gpuDataC)
	{
		gpuDataC=gpuDataP=0;
		primTableJ[gpuCommand]((unsigned char *)gpuData);
	}
	return;
}

// this function will be removed soon
void CALLBACK GPUsetMode(unsigned long gdata)
{
	imageTransfer = gdata;
	return;
}

// this function will be removed soon
long CALLBACK GPUgetMode(void)
{
	return imageTransfer;
}


long CALLBACK GPUconfigure(void)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CCnfDlg myDlg;
	DWORD temp;
	HKEY key;

	myDlg.m_bFullscreen=gpuConfig.bFullscreen;
	myDlg.m_bBilinear=gpuConfig.bBilinear;
	myDlg.m_bTextureHalf=gpuConfig.bTextureHalf;
	myDlg.m_nMaxTextures=gpuConfig.nMaxTextures;
	myDlg.m_bDumpVRAM=gpuConfig.bDumpVRAM;
	myDlg.m_bWireFrame=gpuConfig.bWireFrame;
	myDlg.m_bAntialias=gpuConfig.bAntialias;
	myDlg.m_bClearScreen=gpuConfig.bClearScreen;
	myDlg.m_bCheckPalette=gpuConfig.bCheckPalette;
	if(myDlg.DoModal()==IDOK){
		RegCreateKey(HKEY_CURRENT_USER,"Software\\Vision Thing\\PSEmu Pro\\GPU\\ZenOpenGL",&key);
		temp=gpuConfig.bFullscreen=myDlg.m_bFullscreen;
		RegSetValueEx(key,"FullScreen",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bBilinear=myDlg.m_bBilinear;
		RegSetValueEx(key,"Bilinear",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bTextureHalf=myDlg.m_bTextureHalf;
		RegSetValueEx(key,"TextureHalf",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.nMaxTextures=myDlg.m_nMaxTextures;
		RegSetValueEx(key,"MaxTextures",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bDumpVRAM=myDlg.m_bDumpVRAM;
		RegSetValueEx(key,"DumpVRAM",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bWireFrame=myDlg.m_bWireFrame;
		RegSetValueEx(key,"WireFrame",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bAntialias=myDlg.m_bAntialias;
		RegSetValueEx(key,"Antialias",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bClearScreen=myDlg.m_bClearScreen;
		RegSetValueEx(key,"ClearScreen",0,REG_DWORD,(CONST BYTE *)&temp,4);
		temp=gpuConfig.bCheckPalette=myDlg.m_bCheckPalette;
		RegSetValueEx(key,"CheckPalette",0,REG_DWORD,(CONST BYTE *)&temp,4);
		RegCloseKey(key);
	}
	return 0;
}

long CALLBACK GPUdmaChain(unsigned long * baseAddrL, unsigned long addr)
{
	// leave this function in peace :)
	// this is under high optimizing now
	// and however you dont need to change this
	
	unsigned long dmaMem;
	unsigned char * baseAddrB;
	unsigned char count;
	unsigned char cmd;

	__asm
	{
		push	esi
		push	edi
/*		mov		ebx,baseAddrL
		mov		eax,addr
		movzx	ecx, byte ptr[ebx+eax+3]	// count
dmaCountLoop:
		test	ecx,ecx
		jne		dmaCountLoop
*/
	}

	//syslog(0,"*DMA CHAIN*");
	
	baseAddrB = (unsigned char*) baseAddrL;
	do
	{
		count = baseAddrB[addr+3];
		dmaMem = addr+4;
		while (count)
		{
			cmd = baseAddrB[dmaMem+3];
			//syslog(2,"     %08x %02x",baseAddrL[dmaMem/4],baseAddrB[dmaMem+3]);
			if (primTableC[cmd] == 0)
			{
				if (baseAddrL[dmaMem/4]!=0)
					syslog(255,"DMA_CMD *NI* %04x %04x",dmaMem,baseAddrL[dmaMem/4]);
				dmaMem+=4;
				count--;
			}
			else
			{
//				for (xx=0;xx<primTableC[cmd];xx++)
//					syslog(255,"DMA %04x %04x", dmaMem+xx*4,baseAddrL[dmaMem/4+xx]);
				primTableJ[cmd]((unsigned char*)baseAddrL + dmaMem);
				dmaMem+=(primTableC[cmd]*4);
				count-=primTableC[cmd];
			}
		}
		addr = baseAddrL[addr/4]&0xffffff;
		if (addr <= 0 || addr == 0xffffff) break;;
		//addr&=0x1fffff;
	} while (TRUE);
	__asm
	{
		pop		edi
		pop		esi
	}
	// returns default
	return 0;
}



void CALLBACK GPUabout(void)
{
	// do not edit this function
	// modify only Resource Dialog (IDD_ABTDLG)
	AFX_MANAGE_STATE(AfxGetStaticModuleState());
	CAbtDlg myDlg;
	myDlg.DoModal();
}

long CALLBACK GPUtest(void)
{
	// if test fails this function should return negative value for error (unable to continue)
	// and positive value for warning (can continue but output might be crappy)
	return 0;
}
